var _a;
import { LocalNode } from "cojson";
import { Context } from "effect";
import { CoValueBase, Profile, ValueRef, inspect } from "../internal.js";
export class Account extends CoValueBase {
    get _encoding() {
        return this.constructor._encoding;
    }
    get _owner() {
        return this;
    }
    get _loadedAs() {
        return this.isMe
            ? this
            : _a.fromNode(this._raw.core.node);
    }
    get _refs() {
        const profileID = this._raw.get("profile");
        const rootID = this._raw.get("root");
        return {
            profile: profileID && (new ValueRef(profileID, this._loadedAs, this._encoding.profile.ref())),
            root: rootID &&
                new ValueRef(rootID, this._loadedAs, this._encoding.root.ref()),
        };
    }
    constructor(init, options) {
        super();
        this.acceptInvite = (async (valueID, inviteSecret, coValueClass) => {
            if (!this.isMe) {
                throw new Error("Only a controlled account can accept invites");
            }
            await this._raw.acceptInvite(valueID, inviteSecret);
            return coValueClass.load(valueID, {
                as: this,
            });
        });
        if (!("fromRaw" in options)) {
            throw new Error("Can only construct account from raw or with .create()");
        }
        this.isMe = options.fromRaw.id == options.fromRaw.core.node.account.id;
        Object.defineProperties(this, {
            id: {
                value: options.fromRaw.id,
                enumerable: false,
            },
            _raw: { value: options.fromRaw, enumerable: false },
            profile: {
                get: () => {
                    const ref = this._refs.profile;
                    return ref ? ref.accessFrom(this) : undefined;
                },
                set: (value) => {
                    if (value) {
                        this._raw.set("profile", value.id);
                    }
                },
            },
            root: {
                get: () => {
                    const ref = this._refs.root;
                    return ref ? ref.accessFrom(this) : undefined;
                },
                set: (value) => {
                    if (value) {
                        this._raw.set("root", value.id);
                    }
                },
            },
        });
        if (this.isMe) {
            this.sessionID =
                options.fromRaw.core.node.currentSessionID;
        }
    }
    myRole() {
        if (this.isMe) {
            return "admin";
        }
    }
    static async create(options) {
        const { node } = await LocalNode.withNewlyCreatedAccount({
            ...options,
            migration: async (rawAccount) => {
                const account = new this(undefined, {
                    fromRaw: rawAccount,
                });
                await account.migrate?.();
            },
        });
        return this.fromNode(node);
    }
    static async become(options) {
        const node = await LocalNode.withLoadedAccount({
            accountID: options.accountID,
            accountSecret: options.accountSecret,
            sessionID: options.sessionID,
            peersToLoadFrom: options.peersToLoadFrom,
            migration: async (rawAccount) => {
                const account = new this(undefined, {
                    fromRaw: rawAccount,
                });
                await account.migrate?.();
            },
        });
        return this.fromNode(node);
    }
    static fromNode(node) {
        return new this(undefined, {
            fromRaw: node.account,
        });
    }
    toJSON() {
        return {
            id: this.id,
            _type: this._type,
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    static encoding(fields) {
        this._encoding || (this._encoding = {});
        Object.assign(this._encoding, fields);
    }
}
_a = Account;
(() => {
    _a._encoding = {
        profile: { ref: () => Profile },
        root: { json: true },
    };
})();
export class AccountCtx extends Context.Tag("Account")() {
}
export function isControlledAccount(account) {
    return account.isMe;
}
//# sourceMappingURL=account.js.map