var _a;
import { RawAccount } from "cojson";
import { Account, CoValueBase, Group, ValueRef, inspect, makeRefs, } from "../internal.js";
import { Schema } from "@effect/schema";
export class CoList extends Array {
    get _encoding() {
        return this.constructor._encoding;
    }
    get _owner() {
        return this._raw.group instanceof RawAccount
            ? Account.fromRaw(this._raw.group)
            : Group.fromRaw(this._raw.group);
    }
    get _refs() {
        return makeRefs((idx) => this._raw.get(idx), () => Array.from({ length: this._raw.entries().length }, (_, idx) => idx), this._loadedAs, (_idx) => this._encoding._item.ref());
    }
    get _edits() {
        throw new Error("Not implemented");
    }
    get _loadedAs() {
        return Account.fromNode(this._raw.core.node);
    }
    constructor(init, options) {
        if (typeof init === "number") {
            // this might be called from an intrinsic, like map, trying to create an empty array
            // passing `0` as the only parameter
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            return new Array(init);
        }
        super();
        if (!options) {
            throw new Error("Must provide options");
        }
        let raw;
        if ("fromRaw" in options) {
            raw = options.fromRaw;
        }
        else {
            const rawOwner = options.owner._raw;
            const rawInit = init && this.toRawItems(init);
            raw = rawOwner.createList(rawInit);
        }
        Object.defineProperties(this, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _raw: { value: raw, enumerable: false },
        });
        return new Proxy(this, CoListProxyHandler(this._encoding._item));
    }
    toRawItems(items) {
        const itemDescriptor = this._encoding._item;
        const rawItems = itemDescriptor === "json"
            ? items
            : "encoded" in itemDescriptor
                ? items?.map((e) => Schema.encodeSync(itemDescriptor.encoded)(e))
                : "ref" in itemDescriptor
                    ? items?.map((v) => v.id)
                    : (() => {
                        throw new Error("Invalid element descriptor");
                    })();
        return rawItems;
    }
    push(...items) {
        for (const item of this.toRawItems(items)) {
            this._raw.append(item);
        }
        return this._raw.entries().length;
    }
    unshift(...items) {
        for (const item of this.toRawItems(items)) {
            this._raw.prepend(item);
        }
        return this._raw.entries().length;
    }
    pop() {
        const last = this[this.length - 1];
        this._raw.delete(this.length - 1);
        return last;
    }
    shift() {
        const first = this[0];
        this._raw.delete(0);
        return first;
    }
    splice(start, deleteCount, ...items) {
        const deleted = this.slice(start, start + deleteCount);
        for (let idxToDelete = start + deleteCount; idxToDelete > start; idxToDelete--) {
            this._raw.delete(idxToDelete);
        }
        let appendAfter = start;
        for (const item of this.toRawItems(items)) {
            this._raw.append(item, appendAfter);
            appendAfter++;
        }
        return deleted;
    }
    toJSON() {
        const itemDescriptor = this._encoding._item;
        if (itemDescriptor === "json") {
            return this._raw.asArray();
        }
        else if ("encoded" in itemDescriptor) {
            return this._raw
                .asArray()
                .map((e) => Schema.encodeSync(itemDescriptor.encoded)(e));
        }
        else if ("ref" in itemDescriptor) {
            return this.map((item) => item?.toJSON());
        }
        else {
            return [];
        }
    }
    [inspect]() {
        return this.toJSON();
    }
    static fromRaw(raw) {
        return new this(undefined, { fromRaw: raw });
    }
    static encoding(def) {
        this._encoding || (this._encoding = {});
        Object.assign(this._encoding, def);
    }
}
_a = CoList;
(() => {
    _a.prototype._type = "CoList";
})();
(() => {
    _a.prototype.subscribe = CoValueBase.prototype.subscribe;
})();
(() => {
    _a.prototype.subscribeEf = CoValueBase.prototype.subscribeEf;
})();
CoList.loadEf = CoValueBase.loadEf;
CoList.load = CoValueBase.load;
CoList.subscribeEf = CoValueBase.subscribeEf;
CoList.subscribe = CoValueBase.subscribe;
function CoListProxyHandler(itemDescriptor) {
    return {
        get(target, key, receiver) {
            if (typeof key === "string" && !isNaN(+key)) {
                const rawValue = target._raw.get(Number(key));
                if (itemDescriptor === "json") {
                    return rawValue;
                }
                else if ("encoded" in itemDescriptor) {
                    return rawValue === undefined
                        ? undefined
                        : Schema.decodeSync(itemDescriptor.encoded)(rawValue);
                }
                else if ("ref" in itemDescriptor) {
                    return rawValue === undefined
                        ? undefined
                        : new ValueRef(rawValue, target._loadedAs, itemDescriptor.ref()).accessFrom(receiver);
                }
            }
            else if (key === "length") {
                return target._raw.entries().length;
            }
            else {
                return Reflect.get(target, key, receiver);
            }
        },
        set(target, key, value, receiver) {
            if (typeof key === "string" && !isNaN(+key)) {
                let rawValue;
                if (itemDescriptor === "json") {
                    rawValue = value;
                }
                else if ("encoded" in itemDescriptor) {
                    rawValue = Schema.encodeSync(itemDescriptor.encoded)(value);
                }
                else if ("ref" in itemDescriptor) {
                    rawValue = value.id;
                }
                target._raw.replace(Number(key), rawValue);
                return true;
            }
            else {
                return Reflect.set(target, key, value, receiver);
            }
        },
        has(target, key) {
            if (typeof key === "string" && !isNaN(+key)) {
                return Number(key) < target._raw.entries().length;
            }
            else {
                return Reflect.has(target, key);
            }
        },
    };
}
//# sourceMappingURL=coList.js.map