var _a;
import { Schema } from "@effect/schema";
import { Account, CoValueBase, ValueRef, inspect, makeRefs, subscriptionsScopes, indexSignature, } from "../internal.js";
export class CoMap extends CoValueBase {
    get _encoding() {
        return this.constructor._encoding;
    }
    get _refs() {
        return makeRefs((key) => this._raw.get(key), () => Object.keys(this._encoding).filter((key) => {
            const schema = this._encoding[key];
            schema !== "json" && "ref" in schema;
        }), this._loadedAs, (key) => this._encoding[key].ref());
    }
    get _edits() {
        return new Proxy(this, {
            get(target, key) {
                const rawEdit = target._raw.lastEditAt(key);
                if (!rawEdit)
                    return undefined;
                const descriptor = target._encoding[key];
                return {
                    value: descriptor === "json"
                        ? rawEdit.value
                        : "encoded" in descriptor
                            ? Schema.decodeSync(descriptor.encoded)(rawEdit.value)
                            : new ValueRef(rawEdit.value, target._loadedAs, descriptor.ref()).accessFrom(target),
                    ref: descriptor !== "json" && "ref" in descriptor
                        ? new ValueRef(rawEdit.value, target._loadedAs, descriptor.ref())
                        : undefined,
                    by: rawEdit.by &&
                        new ValueRef(rawEdit.by, target._loadedAs, Account).accessFrom(target),
                    madeAt: rawEdit.at,
                };
            },
        });
    }
    get _loadedAs() {
        return Account.fromNode(this._raw.core.node);
    }
    constructor(init, options) {
        super();
        if (!this._encoding) {
            throw new Error("No schema found in " +
                this.constructor.name +
                " - ensure that you have a `static { this.define({...}) }` block in the class definition.");
        }
        const raw = this.rawFromInit(options, init);
        Object.defineProperties(this, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _raw: { value: raw, enumerable: false },
        });
        this.definePropertiesFromSchema();
        if (this._encoding[indexSignature]) {
            return new Proxy(this, CoMapProxyHandler());
        }
    }
    toJSON() {
        const jsonedFields = this._raw.keys().map((key) => {
            const tKey = key;
            const descriptor = this._encoding[tKey];
            if (descriptor == "json" || "encode" in descriptor) {
                return [key, this._raw.get(key)];
            }
            else if ("ref" in descriptor) {
                const jsonedRef = this[tKey]?.toJSON();
                return [key, jsonedRef];
            }
            else {
                return [key, undefined];
            }
        });
        return {
            id: this.id,
            _type: this._type,
            ...Object.fromEntries(jsonedFields),
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    rawFromInit(options, init) {
        let raw;
        if ("fromRaw" in options) {
            raw = options.fromRaw;
        }
        else {
            const rawOwner = options.owner._raw;
            const rawInit = {};
            if (init)
                for (const key of Object.keys(init)) {
                    const initValue = init[key];
                    const descriptor = (this._encoding[key] || this._encoding[indexSignature]);
                    if (descriptor === "json") {
                        rawInit[key] = initValue;
                    }
                    else if ("ref" in descriptor) {
                        if (initValue) {
                            rawInit[key] = initValue.id;
                        }
                    }
                    else if ("encoded" in descriptor) {
                        rawInit[key] = Schema.encodeSync(descriptor.encoded)(initValue);
                    }
                }
            raw = rawOwner.createMap(rawInit);
        }
        return raw;
    }
    static encoding(fields) {
        this._encoding || (this._encoding = {});
        Object.assign(this._encoding, fields);
    }
    definePropertiesFromSchema() {
        for (const [key, fieldSchema] of Object.entries(this._encoding)) {
            if (key === "indexSignature")
                continue;
            const descriptor = fieldSchema;
            if (descriptor === "json") {
                Object.defineProperty(this, key, this.primitivePropDef(key));
            }
            else if ("encoded" in descriptor) {
                Object.defineProperty(this, key, this.encodedPropDef(key, descriptor.encoded));
            }
            else if ("ref" in descriptor) {
                Object.defineProperty(this, key, this.refPropDef(key, descriptor.ref));
            }
        }
    }
    primitivePropDef(key) {
        return {
            get: () => {
                return this._raw.get(key);
            },
            set(value) {
                this._raw.set(key, value);
            },
            enumerable: true,
            configurable: true,
        };
    }
    encodedPropDef(key, arg) {
        return {
            get: () => {
                const raw = this._raw.get(key);
                return raw === undefined
                    ? undefined
                    : Schema.decodeSync(arg)(raw);
            },
            set(value) {
                this._raw.set(key, Schema.encodeSync(arg)(value));
            },
            enumerable: true,
            configurable: true,
        };
    }
    refPropDef(key, ref) {
        return {
            get: () => {
                const rawID = this._raw.get(key);
                return rawID === undefined
                    ? undefined
                    : new ValueRef(rawID, this._loadedAs, ref()).accessFrom(this);
            },
            set: (value) => {
                this._raw.set(key, value.id);
                subscriptionsScopes.get(this)?.onRefAccessedOrSet(value.id);
            },
            enumerable: true,
            configurable: true,
        };
    }
}
_a = CoMap;
(() => {
    _a.prototype._type = "CoMap";
})();
// TODO: cache handlers per descriptor for performance?
function CoMapProxyHandler() {
    return {
        get(target, key, receiver) {
            const descriptor = target._encoding[indexSignature];
            if (key in target || typeof key === "symbol") {
                return Reflect.get(target, key, receiver);
            }
            else {
                const raw = target._raw.get(key);
                if (descriptor === "json") {
                    return raw;
                }
                else if ("encoded" in descriptor) {
                    return raw === undefined
                        ? undefined
                        : Schema.decodeSync(descriptor.encoded)(raw);
                }
                else if ("ref" in descriptor) {
                    return raw === undefined
                        ? undefined
                        : new ValueRef(raw, target._loadedAs, descriptor.ref()).accessFrom(target);
                }
            }
        },
        set(target, key, value, receiver) {
            const descriptor = target._encoding[indexSignature];
            if (key in target || typeof key === "symbol") {
                return Reflect.set(target, key, value, receiver);
            }
            else {
                if (descriptor === "json") {
                    target._raw.set(key, value);
                }
                else if ("encoded" in descriptor) {
                    target._raw.set(key, Schema.encodeSync(descriptor.encoded)(value));
                }
                else if ("ref" in descriptor) {
                    target._raw.set(key, value.id);
                    subscriptionsScopes
                        .get(target)
                        ?.onRefAccessedOrSet(value.id);
                }
                return true;
            }
        },
        ownKeys(target) {
            const keys = Reflect.ownKeys(target);
            for (const key of target._raw.keys()) {
                if (!keys.includes(key)) {
                    keys.push(key);
                }
            }
            return keys;
        },
        getOwnPropertyDescriptor(target, key) {
            if (key in target) {
                return Reflect.getOwnPropertyDescriptor(target, key);
            }
            else if (key in target._raw.ops) {
                return {
                    enumerable: true,
                    configurable: true,
                    writable: true,
                };
            }
        },
    };
}
//# sourceMappingURL=coMap.js.map