var _a;
import { cojsonInternals } from "cojson";
import { Account, CoValueBase, ValueRef, inspect } from "../internal.js";
import { Schema } from "@effect/schema";
export class CoStream extends CoValueBase {
    get _encoding() {
        return this.constructor._encoding;
    }
    get byMe() {
        return this.by[this._loadedAs.id];
    }
    get inCurrentSession() {
        return this.in[this._loadedAs.sessionID];
    }
    constructor(init, options) {
        super();
        this.by = {};
        this.in = {};
        let raw;
        if ("fromRaw" in options) {
            raw = options.fromRaw;
        }
        else {
            const rawOwner = options.owner._raw;
            raw = rawOwner.createStream();
        }
        Object.defineProperties(this, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _raw: { value: raw, enumerable: false },
        });
        if (init !== undefined) {
            for (const item of init) {
                this.pushItem(item);
            }
        }
        this.updateEntries();
    }
    updateEntries() {
        for (const accountID of this._raw.accounts()) {
            Object.defineProperty(this.by, accountID, {
                get: () => {
                    const rawEntry = this._raw.lastItemBy(accountID);
                    if (!rawEntry)
                        return;
                    return entryFromRawEntry(this, rawEntry, this._loadedAs, accountID, this._encoding._item);
                },
                configurable: true,
                enumerable: true,
            });
        }
        for (const sessionID of this._raw.sessions()) {
            Object.defineProperty(this.in, sessionID, {
                get: () => {
                    const rawEntry = this._raw.lastItemIn(sessionID);
                    if (!rawEntry)
                        return;
                    const by = cojsonInternals.accountOrAgentIDfromSessionID(sessionID);
                    return entryFromRawEntry(this, rawEntry, this._loadedAs, cojsonInternals.isAccountID(by)
                        ? by
                        : undefined, this._encoding._item);
                },
                configurable: true,
                enumerable: true,
            });
        }
    }
    push(...items) {
        for (const item of items) {
            this.pushItem(item);
        }
        this.updateEntries();
    }
    pushItem(item) {
        const itemDescriptor = this._encoding._item;
        if (itemDescriptor === "json") {
            this._raw.push(item);
        }
        else if ("encoded" in itemDescriptor) {
            this._raw.push(Schema.encodeSync(itemDescriptor.encoded)(item));
        }
        else if ("ref" in itemDescriptor) {
            this._raw.push(item.id);
        }
    }
    toJSON() {
        const itemDescriptor = this._encoding._item;
        const mapper = itemDescriptor === "json"
            ? (v) => v
            : "encoded" in itemDescriptor
                ? Schema.encodeSync(itemDescriptor.encoded)
                : (v) => v && v.id;
        return {
            id: this.id,
            _type: this._type,
            by: Object.fromEntries(Object.entries(this.by).map(([account, entry]) => [
                account,
                mapper(entry.value),
            ])),
            in: Object.fromEntries(Object.entries(this.in).map(([session, entry]) => [
                session,
                mapper(entry.value),
            ])),
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    static encoding(def) {
        this._encoding || (this._encoding = {});
        Object.assign(this._encoding, def);
    }
}
_a = CoStream;
(() => {
    _a.prototype._type = "CoStream";
})();
function entryFromRawEntry(accessFrom, rawEntry, loadedAs, accountID, itemField) {
    return {
        get value() {
            if (itemField === "json") {
                return rawEntry.value;
            }
            else if ("encoded" in itemField) {
                return Schema.decodeSync(itemField.encoded)(rawEntry.value);
            }
            else if ("ref" in itemField) {
                return this.ref?.accessFrom(accessFrom);
            }
        },
        get ref() {
            if (itemField !== "json" && "ref" in itemField) {
                const rawId = rawEntry.value;
                return new ValueRef(rawId, loadedAs, itemField.ref());
            }
        },
        get by() {
            return (accountID &&
                new ValueRef(accountID, loadedAs, Account)?.accessFrom(accessFrom));
        },
        madeAt: rawEntry.at,
        tx: rawEntry.tx,
    };
}
export class BinaryCoStream extends CoValueBase {
    constructor(init, options) {
        super();
        let raw;
        if ("fromRaw" in options) {
            raw = options.fromRaw;
        }
        else {
            const rawOwner = options.owner._raw;
            raw = rawOwner.createBinaryStream();
        }
        Object.defineProperties(this, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _raw: { value: raw, enumerable: false },
        });
    }
    getChunks(options) {
        return this._raw.getBinaryChunks(options?.allowUnfinished);
    }
    start(options) {
        this._raw.startBinaryStream(options);
    }
    push(data) {
        this._raw.pushBinaryStreamChunk(data);
    }
    end() {
        this._raw.endBinaryStream();
    }
    toJSON() {
        return {
            id: this.id,
            _type: this._type,
            ...this.getChunks(),
        };
    }
    [inspect]() {
        return this.toJSON();
    }
}
//# sourceMappingURL=coStream.js.map